import sys

import dbt.adapters.factory
from dbt.exceptions import CompilationError
from pydantic import BaseModel

from jinjat.core.exceptions import ExecuteSqlFailure

# This is critical because `get_adapter` is all over dbt-core
# as they expect a singleton adapter instance per plugin,
# so dbt-niceDatabase will have one adapter instance named niceDatabase.
# This makes sense in dbt-land where we have a single Project/Profile
# combination executed in process from start to finish or a single tenant RPC
# This doesn't fit our paradigm of one adapter per DbtProject in a multitenant server,
# so we create an adapter instance **independent** of the FACTORY cache
# and attach it directly to our RuntimeConfig which is passed through
# anywhere dbt-core needs config including in all `get_adapter` calls
dbt.adapters.factory.get_adapter = lambda config: config.adapter

from jinjat.core.dbt.config import ConfigInterface, YamlHandler, JINJAT_REQUEST_VAR_NAME, RAW_CODE, COMPILED_CODE, \
    has_jinja, T
from jinjat.core.models import DbtQueryRequestContext, DbtAdapterExecutionResult, DbtAdapterCompilationResult
import os
import threading
import time
import uuid
from collections import OrderedDict
from copy import copy
from functools import lru_cache
from functools import partial
from typing import (
    Any,
    Callable,
    Tuple,
)
from typing import (
    Dict,
    List,
    Optional,
)

import agate
from dbt.adapters.base import BaseRelation
from dbt.adapters.factory import Adapter, get_adapter_class_by_name
from dbt.config.runtime import RuntimeConfig
from dbt.context.providers import generate_runtime_model_context
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.graph.manifest import ManifestNode, MaybeNonSource, MaybeParsedSource
from dbt.flags import set_from_args
from dbt.node_types import NodeType
from dbt.parser.manifest import ManifestLoader, process_node
from dbt.parser.sql import SqlBlockParser, SqlMacroParser
from dbt.task.parse import MANIFEST_FILE_NAME
from dbt.task.sql import SqlCompileRunner, SqlExecuteRunner

from jinjat.core.log_controller import logger


class DbtProject:
    """Container for a dbt project. The dbt attribute is the primary interface for
    dbt-core. The adapter attribute is the primary interface for the dbt adapter"""

    ADAPTER_TTL = 3600

    def __init__(
            self,
            target: Optional[str] = None,
            profiles_dir: Optional[str] = None,
            project_dir: Optional[str] = None,
            threads: Optional[int] = 1,
            vars: Optional[str] = "{}",
            profile: Optional[str] = None,
    ):
        self.args = ConfigInterface(
            threads=threads,
            target=target,
            profiles_dir=profiles_dir,
            project_dir=project_dir,
            vars=vars,
            profile=profile
        )

        self.parse_project(init=True)

        # Utilities
        self._yaml_handler: Optional[YamlHandler] = None
        self._sql_parser: Optional[SqlBlockParser] = None
        self._macro_parser: Optional[SqlMacroParser] = None
        self._sql_runner: Optional[SqlExecuteRunner] = None
        self._sql_compiler: Optional[SqlCompileRunner] = None

        # Tracks internal state version
        self._version: int = 1
        self.mutex = threading.Lock()
        # atexit.register(lambda dbt_project: dbt_project.adapter.connections.cleanup_all, self)

    def get_adapter(self):
        """This inits a new Adapter which is fundamentally different than
        the singleton approach in the core lib"""
        adapter_name = self.config.credentials.type
        return get_adapter_class_by_name(adapter_name)(self.config)

    def init_adapter(self):
        """Initialize a dbt adapter."""
        if hasattr(self, "_adapter"):
            self._adapter.connections.cleanup_all()
        # The setter verifies connection, resets TTL, and updates adapter ref on config
        self.adapter = self.get_adapter()

    @property
    def adapter(self):
        """dbt-core adapter with TTL and automatic reinstantiation"""
        if time.time() - self._adapter_ttl > self.ADAPTER_TTL:
            logger().info("TTL expired, reinitializing adapter!")
            self.init_adapter()
        return self._adapter

    @adapter.setter
    def adapter(self, adapter: Adapter):
        """Verify connection and reset TTL on adapter set, update adapter prop ref on config"""
        self._adapter = self._verify_connection(adapter)
        self._adapter_ttl = time.time()
        self.config.adapter = self.adapter

    def parse_project(self, init: bool = False) -> None:
        """Parses project on disk from `ConfigInterface` in args attribute, verifies connection
        to adapters database, mutates config, adapter, and dbt attributes"""
        if init:
            set_from_args(self.args, self.args)
            self.config = RuntimeConfig.from_args(self.args)
            self.init_adapter()

        project_parser = ManifestLoader(
            self.config, self.config.load_dependencies(), self.adapter.connections.set_query_header
        )
        # endpatched (https://github.com/dbt-labs/dbt-core/blob/main/core/dbt/parser/manifest.py#L545)
        try:
            self.dbt = project_parser.load()
        except CompilationError as e:
            logger().error(f"Encountered an error loading dbt module:\n{e}")
            sys.exit(1)
        self.dbt.build_flat_graph()
        project_parser.save_macros_to_adapter(self.adapter)
        self._sql_parser = None
        self._macro_parser = None
        self._sql_compiler = None
        self._sql_runner = None

    @property
    def yaml_handler(self) -> YamlHandler:
        """A YAML handler for loading and dumping yaml files on disk"""
        if self._yaml_handler is None:
            self._yaml_handler = YamlHandler()
        return self._yaml_handler

    @property
    def sql_parser(self) -> SqlBlockParser:
        """A dbt-core SQL parser capable of parsing and adding nodes to the manifest via `parse_remote` which will
        also return the added node to the caller. Note that post-parsing this still typically requires calls to
        `_process_nodes_for_ref` and `_process_sources_for_ref` from `dbt.parser.manifest`"""
        if self._sql_parser is None:
            self._sql_parser = SqlBlockParser(self.config, self.dbt, self.config)
        return self._sql_parser

    @property
    def macro_parser(self) -> SqlMacroParser:
        """A dbt-core macro parser"""
        if self._macro_parser is None:
            self._macro_parser = SqlMacroParser(self.config, self.dbt)
        return self._macro_parser

    @property
    def sql_runner(self) -> SqlExecuteRunner:
        """A runner which is used internally by the `execute_sql` function of `dbt.lib`.
        The runners `node` attribute can be updated before calling `compile` or `compile_and_execute`."""
        if self._sql_runner is None:
            self._sql_runner = SqlExecuteRunner(
                self.config, self.adapter, node=None, node_index=1, num_nodes=1
            )
        return self._sql_runner

    @property
    def sql_compiler(self) -> SqlCompileRunner:
        """A runner which is used internally by the `compile_sql` function of `dbt.lib`.
        The runners `node` attribute can be updated before calling `compile` or `compile_and_execute`."""
        if self._sql_compiler is None:
            self._sql_compiler = SqlCompileRunner(
                self.config, self.adapter, node=None, node_index=1, num_nodes=1
            )
        return self._sql_compiler

    def _verify_connection(self, adapter: Adapter) -> Adapter:
        """Verification for adapter + profile. Used as a passthrough,
        ie: `self.adapter = _verify_connection(get_adapter(...))`
        This also seeds the master connection"""
        try:
            adapter.connections.set_connection_name()
            adapter.debug_query()
        except Exception as query_exc:
            raise Exception(f"Could not connect to Database: {query_exc}") from query_exc
        else:
            return adapter

    def adapter_probe(self) -> bool:
        """Check adapter connection, useful for long running processes such as the server or playground"""
        if not hasattr(self, "adapter") or self.adapter is None:
            return False
        try:
            with self.adapter.connection_named("jinjat-heartbeat"):
                self.adapter.debug_query()
        except Exception:
            # TODO: we can decide to reinit the Adapter here
            return False
        logger().info("Heartbeat received for %s", self.project_name)
        return True

    def fn_threaded_conn(self, fn: Callable[..., T], *args, **kwargs) -> Callable[..., T]:
        """Used for jobs which are intended to be submitted to a thread pool,
        the 'master' thread should always have an available connection for the duration of
        typical program runtime by virtue of the `_verify_connection` method.
        Threads however require singleton seeding"""

        def _with_conn() -> T:
            self.adapter.connections.set_connection_name()
            return fn(*args, **kwargs)

        return _with_conn

    def generate_runtime_model_context(self, node: ManifestNode):
        """Wraps dbt context provider"""
        return generate_runtime_model_context(node, self.config, self.dbt)

    @property
    def project_name(self) -> str:
        """dbt project name"""
        return self.config.project_name

    @property
    def project_root(self) -> str:
        """dbt project root"""
        return self.config.project_root

    def safe_parse_project(self, reinit: bool = False) -> None:
        """This is used to reseed the DbtProject safely post-init. This is
        intended for use by the jinjat server"""
        if reinit:
            self.clear_caches()
        _config_pointer = copy(self.config)
        try:
            self.parse_project(init=reinit)
        except Exception as parse_error:
            self.config = _config_pointer
            raise parse_error
        self.write_manifest_artifact()

    def write_manifest_artifact(self) -> None:
        """Write a manifest.json to disk"""
        artifact_path = os.path.join(
            self.config.project_root, self.config.target_path, MANIFEST_FILE_NAME
        )
        self.dbt.write(artifact_path)

    def clear_caches(self) -> None:
        """Clear least recently used caches and reinstantiable container objects"""
        self.get_ref_node.cache_clear()
        self.get_source_node.cache_clear()
        self.get_macro_function.cache_clear()
        self.get_columns.cache_clear()
        self.compile_sql.cache_clear()

    @lru_cache(maxsize=10)
    def get_ref_node(self, target_model_name: str) -> MaybeNonSource:
        """Get a `ManifestNode` from a dbt project model name"""
        return self.dbt.resolve_ref(
            target_model_name=target_model_name,
            target_model_package=None,
            current_project=self.config.project_name,
            node_package=self.config.project_name,
        )

    @lru_cache(maxsize=10)
    def get_source_node(self, target_source_name: str, target_table_name: str) -> MaybeParsedSource:
        """Get a `ManifestNode` from a dbt project source name and table name"""
        return self.dbt.resolve_source(
            target_source_name=target_source_name,
            target_table_name=target_table_name,
            current_project=self.config.project_name,
            node_package=self.config.project_name,
        )

    def get_server_node(self, sql: str, node_name="name"):
        """Get a node for SQL execution against adapter"""
        self._clear_node(node_name)
        sql_node = self.sql_parser.parse_remote(sql, node_name)
        process_node(self.config, self.dbt, sql_node)
        return sql_node

    @lru_cache(maxsize=10)
    def get_node_by_path(self, path: str):
        """Find an existing node given relative file path."""
        for node in self.dbt.nodes.values():
            if node.original_file_path == path:
                return node
        return None

    @lru_cache(maxsize=100)
    def get_macro_function(self, macro_name: str) -> Callable[[Dict[str, Any]], Any]:
        """Get macro as a function which takes a dict via argument named `kwargs`,
        ie: `kwargs={"relation": ...}`

        make_schema_fn = get_macro_function('make_schema')\n
        make_schema_fn({'name': '__test_schema_1'})\n
        make_schema_fn({'name': '__test_schema_2'})"""
        return partial(self.adapter.execute_macro, macro_name=macro_name, manifest=self.dbt)

    def adapter_execute(
            self, sql: str, auto_begin: bool = False, fetch: bool = False
    ) -> Tuple[AdapterResponse, agate.Table]:
        """Wraps adapter.execute. Execute SQL against database"""
        return self.adapter.execute(sql, auto_begin, fetch)

    def execute_macro(
            self,
            macro: str,
            kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Wraps adapter execute_macro. Execute a macro like a function."""
        return self.get_macro_function(macro)(kwargs=kwargs)

    def execute_sql(self, raw_sql: str, ctx: DbtQueryRequestContext, fetch: bool = True) -> DbtAdapterExecutionResult:
        """Execute dbt SQL statement against database"""
        # if no jinja chars then these are synonymous
        compiled_sql = raw_sql
        if has_jinja(raw_sql):
            # jinja found, compile it
            try:
                compiled_node = self.compile_sql(raw_sql, ctx)
                compiled_sql = compiled_node.compiled_sql
            except Exception as e:
                raise ExecuteSqlFailure(raw_sql, None, e)
        try:
            table = self.adapter_execute(compiled_sql, fetch=fetch)
        except Exception as e:
            raise ExecuteSqlFailure(raw_sql, compiled_sql, e)

        return DbtAdapterExecutionResult(
            *table,
            raw_sql,
            compiled_sql,
        )

    def compile_sql(self, raw_sql: str, ctx: Optional[DbtQueryRequestContext] = None, retry: int = 3) -> DbtAdapterCompilationResult:
        """Creates a node with `get_server_node` method. Compile generated node.
        Has a retry built in because even uuidv4 cannot gaurantee uniqueness at the speed
        in which we can call this function concurrently. A retry significantly increases the stability"""
        temp_node_id = str(uuid.uuid4())
        try:
            node = self.compile_node(self.get_server_node(raw_sql, temp_node_id), ctx)
        except Exception as exc:
            if retry > 0:
                return self.compile_sql(raw_sql, ctx, retry - 1)
            raise exc
        else:
            return node
        finally:
            self._clear_node(temp_node_id)

    def compile_node(self, node: ManifestNode, ctx: Optional[DbtQueryRequestContext]) -> DbtAdapterCompilationResult:
        """Compiles existing node."""
        self.sql_compiler.node = node
        compiler = self.adapter.get_compiler()
        compiled_node = compiler.compile_node(self.sql_compiler.node,
                                              self.dbt,
                                              {JINJAT_REQUEST_VAR_NAME: ctx} if ctx is not None else {},
                                              write=False)
        return DbtAdapterCompilationResult(
            getattr(compiled_node, RAW_CODE),
            getattr(compiled_node, COMPILED_CODE),
            compiled_node,
        )

    def _clear_node(self, name="name"):
        """Removes the statically named node created by `execute_sql` and `compile_sql` in `dbt.lib`"""
        self.dbt.nodes.pop(f"{NodeType.SqlOperation}.{self.project_name}.{name}", None)

    def get_relation(self, database: str, schema: str, name: str) -> Optional[BaseRelation]:
        """Wrapper for `adapter.get_relation`"""
        return self.adapter.get_relation(database, schema, name)

    def create_relation(self, database: str, schema: str, name: str) -> BaseRelation:
        """Wrapper for `adapter.Relation.create`"""
        return self.adapter.Relation.create(database, schema, name)

    def create_relation_from_node(self, node: ManifestNode) -> BaseRelation:
        """Wrapper for `adapter.Relation.create_from`"""
        return self.adapter.Relation.create_from(self.config, node)

    def get_columns_in_relation(self, node: ManifestNode) -> List[str]:
        """Wrapper for `adapter.get_columns_in_relation`"""
        return self.adapter.get_columns_in_relation(self.create_relation_from_node(node))

    def get_or_create_relation(
            self, database: str, schema: str, name: str
    ) -> Tuple[BaseRelation, bool]:
        """Get relation or create if not exists. Returns tuple of relation and
        boolean result of whether it existed ie: (relation, did_exist)"""
        ref = self.get_relation(database, schema, name)
        return (ref, True) if ref else (self.create_relation(database, schema, name), False)

    def create_schema(self, node: ManifestNode):
        """Create a schema in the database"""
        return self.execute_macro(
            "create_schema",
            kwargs={"relation": self.create_relation_from_node(node)},
        )

    def materialize(
            self, node: ManifestNode, temporary: bool = True
    ) -> Tuple[AdapterResponse, None]:
        """Materialize a table in the database"""
        return self.adapter_execute(
            # Returns CTAS string so send to adapter.execute
            self.execute_macro(
                "create_table_as",
                kwargs={
                    "sql": getattr(node, COMPILED_CODE),
                    "relation": self.create_relation_from_node(node),
                    "temporary": temporary,
                },
            ),
            auto_begin=True,
        )

    @classmethod
    def from_args(cls, args: ConfigInterface) -> "DbtProject":
        """Instatiate the DbtProject directly from a ConfigInterface instance"""
        return cls(
            target=args.target,
            profiles_dir=args.profiles_dir,
            project_dir=args.project_dir,
            threads=args.threads,
        )


class DbtTarget(BaseModel):
    target: Optional[str] = None
    profiles_dir: Optional[str] = None
    project_dir: Optional[str] = None
    refine: Optional[bool] = False
    threads: Optional[int] = 1
    vars: Optional[str] = "{}"


class DbtProjectContainer:
    """This class manages multiple DbtProjects which each correspond
    to a single dbt project on disk. This is mostly for jinjat server use"""

    def __init__(self):
        self._projects: Dict[str, DbtProject] = OrderedDict()
        self._default_project: Optional[str] = None

    def get_project(self, project_name: str = None) -> Optional[DbtProject]:
        """Primary interface to get a project and execute code"""
        if project_name is None:
            return self.get_default_project()
        else:
            return self._projects.get(project_name)

    @lru_cache(maxsize=10)
    def get_project_by_root_dir(self, root_dir: str) -> Optional[DbtProject]:
        """Get a project by its root directory."""
        root_dir = os.path.abspath(os.path.normpath(root_dir))
        for project in self._projects.values():
            if os.path.abspath(project.project_root) == root_dir:
                return project
        return None

    def get_default_project(self) -> Optional[DbtProject]:
        """Gets the default project which at any given time is the
        earliest project inserted into the container"""
        return self._projects.get(self._default_project)

    def add_project(
            self,
            dbt_target: DbtTarget,
            name_override: Optional[str] = None
    ) -> DbtProject:
        """Add a DbtProject with arguments"""
        project = DbtProject(dbt_target.target, dbt_target.profiles_dir, dbt_target.project_dir, dbt_target.threads,
                             dbt_target.vars)
        project_name = name_override or project.config.project_name
        if self._default_project is None:
            self._default_project = project_name
        self._projects[project_name] = project
        return project

    def add_parsed_project(self, project: DbtProject) -> DbtProject:
        """Add an already instantiated DbtProject"""
        self._projects.setdefault(project.config.project_name, project)
        return project

    def add_project_from_args(self, args: ConfigInterface) -> DbtProject:
        """Add a DbtProject from a ConfigInterface"""
        project = DbtProject.from_args(args)
        self._projects.setdefault(project.config.project_name, project)
        return project

    def drop_project(self, project_name: str) -> None:
        """Drop a DbtProject"""
        project = self.get_project(project_name)
        if project is None:
            return
        project.clear_caches()
        project.adapter.connections.cleanup_all()
        self._projects.pop(project_name)
        if self._default_project == project_name:
            if len(self) > 0:
                self._default_project = self._projects.keys()[0]
            else:
                self._default_project = None

    def drop_all_projects(self) -> None:
        """Drop all DbtProjectContainers"""
        self._default_project = None
        for project in self._projects:
            self.drop_project(project)

    def reparse_all_projects(self) -> None:
        """Reparse all projects"""
        for project in self:
            project.safe_parse_project()

    def registered_projects(self) -> List[str]:
        """Convenience to grab all registered project names"""
        return list(self._projects.keys())

    def __len__(self):
        """Allows len(DbtProjectContainer)"""
        return len(self._projects)

    def __getitem__(self, project: str):
        """Allows DbtProjectContainer['jaffle_shop']"""
        maybe_project = self.get_project(project)
        if maybe_project is None:
            raise KeyError(project)
        return maybe_project

    def __delitem__(self, project: str):
        """Allows del DbtProjectContainer['jaffle_shop']"""
        self.drop_project(project)

    def __iter__(self):
        """Allows project for project in DbtProjectContainer"""
        for project in self._projects:
            yield self.get_project(project)

    def __contains__(self, project):
        """Allows 'jaffle_shop' in DbtProjectContainer"""
        return project in self._projects

    def __repr__(self):
        """Canonical string representation of DbtProjectContainer instance"""
        return "\n".join(
            f"Project: {project.project_name}, Dir: {project.project_root}" for project in self
        )

    def __call__(self) -> "DbtProjectContainer":
        """This allows the object to be used as a callable, primarily for FastAPI dependency injection
        ```python
        dbt_project_container = DbtProjectContainer()
        def register(x_dbt_project: str = Header(default=None)):
            dbt_project_container.add_project(...)
        def compile(x_dbt_project: str = Header(default=None), dbt = Depends(dbt_project_container), request: fastapi.Request):
            query = request.body()
            dbt.get_project(x_dbt_project).compile(query)
        ```
        """
        return self



import os
import sys

from jinjat.core.dbt.dbt_project import DbtTarget, DbtProjectContainer
from pathlib import Path
from jinjat.core.log_controller import logger

space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


def print_tree(dir_path: Path, prefix: str = ''):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """
    contents = list(dir_path.iterdir())
    print(contents)
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from print_tree(path, prefix=prefix + extension)


def compile_macro(dbt_target: DbtTarget, macro_name: str, macro_args: str, dry_run: bool = False):
    container = DbtProjectContainer()
    container.add_project(dbt_target)
    project = container.get_default_project()
    macro_result = project.execute_macro(macro_name, macro_args)

    if type(macro_result) != dict:
        logger().fatal(
            "Generator macro needs to return dict[file_path, file_context] but it returned %s the following output: %s",
            type(macro_result), macro_result)
        sys.exit(1)

    files = [(os.path.join(dbt_target.project_dir, path), content) for path, content in macro_result.items()]

    if len(files) == 0:
        logger().info("Macro returned empty object, exiting..")
        sys.exit(0)

    logger().info("Project directory is %s, macro returned %s files to generate", dbt_target.project_dir, len(files))

    if dry_run:
        sys.exit(0)

    dir_structure = generate_dir_structure(list(macro_result.keys()))
    print_dir_structure(dbt_target.project_dir, dir_structure)

    number_of_existing_files = sum(os.path.exists(file[0]) for file in files)
    if number_of_existing_files > 0:
        prompt_message = f"There are {number_of_existing_files} files that you will be overwritten. Type '{number_of_existing_files}' and enter to continue"
        check = str(number_of_existing_files)
    else:
        prompt_message = f"{len(files)} files will be created. Type enter to continue"
        check = ""

    user_output = input(f'{prompt_message}\n')
    if not user_output == check:
        logger().info("Exiting without doing any action.")
        sys.exit(0)

    for (path, content) in files:
        with open(path, "w") as f:
            f.write(content)


# ChatGTP generated the following code
def generate_dir_structure(file_paths):
    dir_structure = {}
    for path in file_paths:
        curr_dir = dir_structure
        components = path.split('/')
        for i, component in enumerate(components[:-1]):
            if component not in curr_dir:
                curr_dir[component] = {}
            curr_dir = curr_dir[component]
            if i == len(components) - 2:
                curr_dir[components[-1]] = path
    return dir_structure


def print_dir_structure(project_dir: str, dir_structure, indent=''):
    # Define the characters used to create the tree-like output
    branch_char = '├── '
    last_branch_char = '└── '
    indent_char = '│   '
    last_indent_char = '    '
    # Iterate over each directory or file in the current directory
    for i, (name, contents) in enumerate(dir_structure.items()):
        # Determine the character to use for the current directory or file
        if i == len(dir_structure) - 1:
            curr_branch_char = last_branch_char
        else:
            curr_branch_char = branch_char
        # Print the current directory or file
        if contents is not None and type(contents) != str:
            print(indent + curr_branch_char + name)
            if i == len(dir_structure) - 1:
                sub_indent = indent + last_indent_char
            else:
                sub_indent = indent + indent_char
            print_dir_structure(project_dir, contents, indent=sub_indent)
        else:
            if os.path.exists(os.path.join(project_dir, contents)):
                sign = '❗️'
            else:
                sign = '✔️'
            print(indent + curr_branch_char + name + ' ' + sign)



from pathlib import Path

from watchdog.events import LoggingEventHandler, PatternMatchingEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver


def watch(directory: str, callback):
    class DbtProjectWatcher(LoggingEventHandler,
                            PatternMatchingEventHandler(patterns=['analysis/*.sql',
                                                                  'analysis/*.yml',
                                                                  'macros/*.sql',
                                                                  'models/*.sql',
                                                                  'dbt_project.yml'])):

        def on_any_event(self, event):
            super().on_any_event(event)
            callback(directory)

    observer = PollingObserver()
    observer.schedule(DbtProjectWatcher, directory, recursive=True)
    observer.start()

def get_project_root() -> Path:
    # TODO: What's the best way?
    return Path(__file__).parent.parent.parent.parent.parent



import asyncio
import functools
import os

import yaml
from deepmerge import always_merger
from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

from jinjat.core.dbt.dbt_project import DbtProjectContainer, DbtProject, DbtTarget
from jinjat.core.log_controller import logger
from jinjat.core.models import JinjatProjectConfig
from jinjat.core.routes.admin import app as admin_app
from jinjat.core.routes.analysis import create_analysis_apps
from jinjat.core.util.api import register_jsonapi_exception_handlers, rapidoc_html, CustomButton, DBT_PROJECT_HEADER, \
    DBT_PROJECT_NAME
from jinjat.core.util.filesystem import get_project_root

SERVER_OPT = "SERVER_OPT"

app = FastAPI(redoc_url=None, docs_url=None)

dbt_container = DbtProjectContainer()
app.state.dbt_project_container = dbt_container
admin_app.state.dbt_project_container = dbt_container


@app.middleware("http")
async def add_cookie_header_for_cors(request: Request, call_next):
    # HACK FOR CorsMiddleware TO BE TRIGGERED
    if request.headers.get('cookie') is None:
        request.headers.__dict__["_list"].append(
            (
                "cookie".encode(),
                f"".encode(),
            )
        )
    response = await call_next(request)
    return response


def get_jinjat_project_config(project_root: str) -> JinjatProjectConfig:
    jinjat_config_file_path = os.path.join(project_root, 'jinjat_project.yml')
    if os.path.exists(jinjat_config_file_path):
        with open(jinjat_config_file_path, 'r') as file_handler:
            jinjat_config_file = file_handler.read()
            try:
                return JinjatProjectConfig.parse_obj(yaml.load(jinjat_config_file, Loader=yaml.Loader))
            except yaml.parser.ParserError:
                raise Exception(f"String '{jinjat_config_file}' is not valid YAML")
    return JinjatProjectConfig()


def unmount_app(new_app: FastAPI):
    existing_apps = next(filter(lambda r: isinstance(r, Mount) and r.name == new_app.root_path, app.routes), None)
    if existing_apps is not None:
        app.routes.remove(existing_apps)


"""Register default project, ugly (using envs?) but works. This will parse the project on disk and load it into memory"""
if SERVER_OPT in os.environ:
    dbt_target = DbtTarget.parse_raw(os.environ[SERVER_OPT])
    logger().info("Registering project: {}".format(dbt_target.json()))
    project = app.state.dbt_project_container.add_project(dbt_target)

    config = get_jinjat_project_config(project.project_root)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=[
            config.cors.allowed_origins
        ],
        allow_credentials=False,
        allow_methods=["*"],
        allow_headers=["*"],
        expose_headers=[DBT_PROJECT_HEADER, DBT_PROJECT_NAME]
    )


    def custom_openapi():
        if app.openapi_schema:
            return app.openapi_schema

        openapi_schema = get_openapi(title=project.project_name, version=project.config.version, routes=app.routes)
        app.openapi_schema = openapi_schema
        if config.openapi is not None:
            always_merger.merge(app.openapi_schema.get('info'), config.openapi.get("info", {}))
        return app.openapi_schema


    register_jsonapi_exception_handlers(app)

    app.openapi = custom_openapi

    analysis_app = create_analysis_apps(config, project)

    logger().info("{} analysis found with `jinjat` config\n".format(len(analysis_app.routes)))

    admin_app.router.add_route("/docs",
                               functools.partial(rapidoc_html,
                                                 CustomButton("Analysis APIs", f"/{project.config.version}/docs")),
                               include_in_schema=False)
    admin_app.version = project.config.version
    app.mount("/admin", admin_app)

    if len(analysis_app.routes) > 0:
        unmount_app(analysis_app)
        app.mount(f"/{project.config.version}", analysis_app)
        app.mount(f"/current", analysis_app)

    default_refine_project = os.path.join(get_project_root(), *["src", "jinjat", "jinjat-refine"])
    project_static_files = os.path.join(project.project_root, "static")

    static_files = main_directory = None
    if os.path.exists(project_static_files) and dbt_target.refine:
        static_files = StaticFiles(directory=project_static_files, html=True)
        static_files.all_directories = [project_static_files, default_refine_project]
        main_directory = default_refine_project
    else:
        if os.path.exists(project_static_files):
            main_directory = project_static_files
        elif dbt_target.refine:
            main_directory = default_refine_project

        if main_directory is not None:
            static_files = StaticFiles(directory=main_directory, html=True)

    if static_files is not None:
        app.mount("/", static_files, name="static")

    if main_directory is None or not os.path.exists(os.path.join(main_directory, "index.html")):
        app.router.add_route("/", lambda request: JSONResponse({
            "analysis_api_docs": f"{request.base_url}{project.config.version}/docs" if len(
                analysis_app.routes) > 0 else None,
            "admin_api_docs": f"{request.base_url}admin/docs",
            "magic": "https://jin.jat",
            "options": dbt_target.dict()
        }))


# We use ambient reinitialization based on TTL now
# loop = asyncio.get_running_loop()
# project.heartbeat = loop.create_task(_adapter_heartbeat(project))
async def _adapter_heartbeat(runner: DbtProject):
    """Equivalent of a keepalive for adapters such as Snowflake"""
    await asyncio.sleep(60 * 30)
    while runner.adapter_probe():
        await asyncio.sleep(60 * 30)



import functools
import json
import re
from copy import deepcopy
from pathlib import Path
from typing import List, Optional, Callable

import jmespath
import jsonref
from dbt.node_types import NodeType
from deepmerge import always_merger
from fastapi import FastAPI
from fastapi.openapi.models import Schema, Response as APIResponse
from fastapi.openapi.utils import get_openapi
from openapi_schema_pydantic import Parameter
from pydantic import ValidationError
from starlette.requests import Request
from starlette.responses import JSONResponse

from jinjat.core.dbt.dbt_project import DbtProject
from jinjat.core.exceptions import InvalidJinjaConfig
from jinjat.core.models import generate_dbt_context_from_request, JinjatExecutionResult, JinjatAnalysisConfig, \
    JinjatProjectConfig, JSON_COLUMNS_QUERY_PARAM
from jinjat.core.routes.admin import _execute_jinjat_query
from jinjat.core.util.api import get_human_readable_error, rapidoc_html, register_jsonapi_exception_handlers, \
    CustomButton
from jinjat.core.util.jmespath import extract_jmespath

ANALYSIS_FILE_PATH_REGEX = re.compile(r"^analysis\/(.*)\.sql$")


async def handle_analysis_api(project: DbtProject,
                              sql: str,
                              openapi: dict,
                              transform_request: Callable[[dict], dict],
                              transform_response: Callable[[dict], dict],
                              fetch: bool,
                              jinjat_project: JinjatProjectConfig,
                              request: Request):
    context = await generate_dbt_context_from_request(request, openapi, transform_request)
    query_result = await _execute_jinjat_query(project, project.execute_sql, sql, context,
                                               request.query_params.get('_limit'), fetch)
    json_cols = json.loads(request.query_params.get(JSON_COLUMNS_QUERY_PARAM) or '[]')
    jinjat_result = JinjatExecutionResult.from_dbt(context, query_result, transform_response, json_cols)

    if context.is_debug_enabled():
        return jinjat_result
    else:
        return jinjat_result.data


def create_components_from_nodes(project: DbtProject):
    schema_nodes = filter(
        lambda node: node.resource_type in ['model', 'seed', 'source'] and 'jinjat' in node.meta,
        project.dbt.nodes.values())

    components = {}
    for node in schema_nodes:
        jinjat = node.config.meta.get('jinjat')
        if jinjat is not None and 'schema' in jinjat:
            schema = Schema.parse_obj(jinjat.get('schema'))
        else:
            schema = Schema(type='object')
        schema.properties = {}
        for column in node.columns.values():
            openapi = column.meta.get('jinjat', {}).get('schema')
            if openapi is None:
                col_schema = Schema()
            else:
                col_schema = Schema.parse_obj(openapi)
            schema.properties[column.name] = col_schema
        components[node.unique_id] = schema.dict(exclude_unset=True)
    return components


def parse_param(path: str) -> (str, List[str]):
    paths = list(Path(ANALYSIS_FILE_PATH_REGEX.findall(path)[0]).parts)
    if paths[-1].startswith('_'):
        paths.pop()
    param_list = []
    for idx, path in enumerate(paths):
        if path.startswith('_'):
            path_name = path[1:]
            paths[idx] = f'{{{path_name}}}'
            param_list.append(path_name)
    return '/'.join(paths), param_list


def compile_transform(jmespath_exp: Optional[str]):
    if jmespath_exp is None or len(jmespath_exp) == 0:
        return lambda body: body

    compiled_transforms = jmespath.compile(jmespath_exp)

    def transform_data(body: dict):
        return compiled_transforms.search(body)

    return compiled_transforms.search


def create_analysis_apps(jinjat_project_config: JinjatProjectConfig, project: DbtProject) -> FastAPI:
    analysis_nodes = filter(lambda node: node.resource_type == NodeType.Analysis.value and node.language == 'sql',
                            project.dbt.nodes.values())
    api = FastAPI(redoc_url=None, docs_url=None, openapi_url=None)
    register_jsonapi_exception_handlers(api)
    api.add_route("/docs", functools.partial(rapidoc_html, CustomButton("Admin APIs", "/admin/docs")),
                  include_in_schema=False)

    # api.add_route("/elements", functools.partial(elements_html, CustomButton("Admin APIs", "/")),
    #               include_in_schema=False)

    async def custom_openapi(req: Request) -> JSONResponse:
        extract_path = req.query_params.get("jmespath")
        if api.openapi_schema:
            return JSONResponse(extract_jmespath(extract_path, api.openapi_schema, project))

        openapi_schema = get_openapi(title=project.project_name, version=project.config.version, routes=api.routes,
                                     servers=[{"url": req.scope.get("root_path", "").rstrip("/")}])

        component_schemas = create_components_from_nodes(project)
        components = openapi_schema.setdefault('components', {})
        existing_schemas = components.setdefault('schemas', {})
        components['schemas'] = {**existing_schemas, **component_schemas}

        if jinjat_project_config.openapi is not None:
            always_merger.merge(openapi_schema, jinjat_project_config.openapi)

        openapi_schema = jsonref.replace_refs(deepcopy(openapi_schema), base_uri="", proxies=False, lazy_load=False)
        api.openapi_schema = openapi_schema
        return JSONResponse(extract_jmespath(extract_path, api.openapi_schema, project))

    api.add_route("/openapi.json", custom_openapi, include_in_schema=False)

    for node in analysis_nodes:
        if 'jinjat' not in node.config.extra:
            jinjat_config = JinjatAnalysisConfig()
        else:
            try:
                jinjat_config = JinjatAnalysisConfig.parse_obj(node.config.extra['jinjat'])
            except ValidationError as e:
                raise InvalidJinjaConfig(node.patch_path, node.original_file_path, get_human_readable_error(e))
        sql = node.raw_code

        methods = [jinjat_config.method] if jinjat_config.method is not None else None
        openapi = jinjat_config.openapi
        fetch_enabled = jinjat_config.fetch

        if not jinjat_config.fetch:
            if openapi.responses is not None:
                raise Exception(
                    "{}: If `fetch` is false, the jinjat.openapi.responses must not set".format(node.unique_id))
            openapi.responses = {204: APIResponse(description="Success")}

        api_path, param_list = parse_param(node.path)

        for param in param_list:
            if openapi.parameters is None:
                openapi.parameters = []
            openapi.parameters.append(Parameter(param_in='path', name=param, required=True,
                                                schema=Schema(type='string').dict(by_alias=True,
                                                                                  exclude_none=True)))

        api_path = f'/{node.package_name}/{api_path}'
        # ctx = generate_parser_model_context(node, project.config, project.dbt,
        #                                     ContextConfig(project.config, node.fqn, node.resource_type,
        #                                                   project.project_name, ))
        # environment = get_environment(node=node, capture_macros=True)
        # inferred_schema = infer_from_ast(environment.from_string(sql, globals=ctx), ignore_constants=True)
        # inferred_schema_json = to_json_schema(inferred_schema)
        # if openapi.requestBody.content.get('application/json').schema()

        openapi_dict = openapi.dict(by_alias=True, exclude_none=True, exclude_unset=True)
        try:
            transform_request = compile_transform(jinjat_config.transform_request)
        except Exception as e:
            raise InvalidJinjaConfig(node.original_file_path, None, f"Unable to parse `transform_request` jmespath expression {jinjat_config.transform_request}: {e}")

        try:
            transform_response = compile_transform(jinjat_config.transform_response)
        except Exception as e:
            raise InvalidJinjaConfig(node.original_file_path, None, f"Unable to parse `transform_response` jmespath expression {jinjat_config.transform_response}: {e}")

        api.add_api_route(api_path,
                          endpoint=functools.partial(handle_analysis_api, project, sql, openapi_dict,
                                                     transform_request,
                                                     transform_response,
                                                     fetch_enabled,
                                                     jinjat_project_config),
                          tags=node.tags,
                          description=node.description,
                          summary=node.name,
                          methods=methods,
                          operation_id=node.unique_id,
                          name=node.alias,
                          openapi_extra=openapi_dict)
    return api



from jsonschema import Draft202012Validator, validators


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        for error in validate_properties(
                validator, properties, instance, schema,
        ):
            yield error

    return validators.extend(
        validator_class, {"properties": set_defaults},
    )


DefaultValidatingValidator = extend_with_default(Draft202012Validator)



import argparse
import json
import os
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Optional

import dbt.config.profile as dbt_profile
import pandas as pd
# import pandas_profiling
import streamlit as st
from dbt.exceptions import CompilationException, DatabaseException
from streamlit_ace import st_ace
# from streamlit_pandas_profiling import st_profile_report

from jinjat.core.dbt.config import DEFAULT_PROFILES_DIR
from jinjat.core.dbt.dbt_project import DbtProject
from jinjat.core.models import DbtQueryRequestContext

st.set_page_config(page_title="Jinjat Playground", page_icon="🌊", layout="wide")
state = st.session_state

try:
    parser = argparse.ArgumentParser(description="Jinjat playground")
    parser.add_argument("--profiles-dir", help="dbt profile directory")
    parser.add_argument("--project-dir", help="dbt project directory")
    args = vars(parser.parse_args(sys.argv[1:]))
except:
    args = {}

root_path = Path(__file__).parent
demo_dir = root_path / "demo"

# GLOBAL STATE VARS
DBT = "DBT"
"""DbtProject object"""
PROJ_DIR = "PROJ_DIR"
"""dbt project directory"""
PROF_DIR = "PROF_DIR"
"""dbt profile directory"""

_proj_dir = args.get("project_dir")
state.setdefault(PROJ_DIR, _proj_dir or os.getenv("DBT_PROJECT_DIR", str(Path.cwd())))
_prof_dir = args.get("profiles_dir")
state.setdefault(PROF_DIR, _prof_dir or os.getenv("DBT_PROFILES_DIR", DEFAULT_PROFILES_DIR))

RAW_PROFILES = "RAW_PROFILES"
"""All profiles as parsed from raw profiles yaml"""
state.setdefault(RAW_PROFILES, dbt_profile.read_profile(state[PROF_DIR] or DEFAULT_PROFILES_DIR))

# SQL PLAYGROUND VARS
SQL_RESULT = "SQL_RESULT"
"""SQL result as a pandas dataframe"""
SQL_ADAPTER_RESP = "SQL_ADAPTER_RESP"
"""Adapter response from dbt"""
SQL_QUERY_STATE = "SQL_QUERY_STATE"
"""SQL query state tracking if it is successful or failed"""

state.setdefault(SQL_RESULT, pd.DataFrame())
state.setdefault(SQL_ADAPTER_RESP, None)
state.setdefault(SQL_QUERY_STATE, "test")

# PRIMARY SQL CONTAINERS
COMPILED_SQL = "COMPILED_SQL"
"""Compiled sql container"""
state.setdefault(COMPILED_SQL, "")

RAW_SQL = "RAW_SQL"
"""Raw sql container"""

state.setdefault(
    RAW_SQL,
    """
{{
    config(
        jinjat={
            "method": "POST"
        }
    )
}}


{% set payment_methods = ['credit_card', 'coupon', 'bank_transfer', 'gift_card'] %}    
{%- set payload = request().body %}

select '{{ payment_methods[payload.payment_method_index] }}'
""".strip(),
)

FLIGHT_REQUEST = "REQUEST_BODY"
"""Request json"""

state.setdefault(FLIGHT_REQUEST, """{
  "method": "POST",
  "headers": {
    "Content-Type": "application/json"
  },
  "body": {
    "payment_method_index": 0
  },
  "query": {
  },
  "params": {
  }
}""")

# COMPONENT KEYS
PROFILE_SELECTOR = "PROFILE_SELECTOR"
"""Selected profile"""
DIALECT_PICKER = "DIALECT_PICKER"
"""Selected SQL dialect for playground"""
QUERY_LIMITER = "QUERY_LIMITER"
"""Limit results returned in SQL runner"""
BASIC_PROFILE_OPT = "BASIC_PROFILE_OPT"
"""Use basic profiling for pandas-profiling"""
PROFILE_DOWNLOADER = "PROFILE_DOWNLOADER"
"""Controller for downloading HTML results of pandas-profiler"""
DYNAMIC_COMPILATION = "DYNAMIC_COMPILATION"
"""Toggle to compile on-type or compile on control+enter"""

# ACE COMPONENT OPTIONS
DBT_ADAPTER_ACE_DIALECT_MAP = {
    "postgres": "pgsql",
    "mysql": "mysql",
    "sqlserver": "sqlserver"
}
DEFAULT_ACE_DIALECT = "sql"

# TRIGGERS
DBT_DO_RELOAD = "DBT_DO_RELOAD"
"""This triggers dbt to reparse the project"""
PIVOT_LAYOUT = "PIVOT_LAYOUT"
"""Pivot the editor layout from side-by-side to top-bottom"""

state.setdefault(PIVOT_LAYOUT, False)

TARGET_PROFILE = "TARGET_PROFILE"
"""Target profile for dbt to execute against"""


def inject_dbt(change_target: Optional[str] = None):
    """Parse dbt project and load context var"""
    if DBT not in state or change_target:
        dbt_ctx = DbtProject(
            project_dir=state[PROJ_DIR],
            profiles_dir=state[PROF_DIR],
            target=change_target,
        )
    else:
        dbt_ctx: DbtProject = state[DBT]
        dbt_ctx.rebuild_dbt_manifest(reset=True)
    state[DBT] = dbt_ctx
    return True


if DBT not in state:
    inject_dbt()
ctx: DbtProject = state[DBT]

state.setdefault(TARGET_PROFILE, ctx.config.target_name)


def toggle_viewer() -> None:
    state[PIVOT_LAYOUT] = not state[PIVOT_LAYOUT]


# @st.cache
def compile_sql(sql: str, _request_context: DbtQueryRequestContext) -> str:
    try:
        return ctx.compile_sql(sql, _request_context).compiled_sql
    except CompilationException:
        return None


def run_query(sql: str, limit: int = 2000) -> None:
    try:
        result = ctx.execute_sql(f"select * from ({sql}) as __all_data limit {limit}")
    except DatabaseException as error:
        state[SQL_QUERY_STATE] = "error"
        state[SQL_ADAPTER_RESP] = str(error)
    else:
        output = [OrderedDict(zip(result.table.column_names, row)) for row in result.table.rows]
        state[SQL_RESULT] = pd.DataFrame(output)
        state[SQL_ADAPTER_RESP] = result.adapter_response
        state[SQL_QUERY_STATE] = "success"

st.sidebar.title("Jinjat")

st.sidebar.button("Switch {} layout".format("horizontal" if state[PIVOT_LAYOUT] else "vertical"),
                  on_click=toggle_viewer)

st.sidebar.header("Profiles")
state[TARGET_PROFILE] = st.sidebar.radio(
    f"Loaded profiles from {ctx.config.profile_name}",
    [target for target in state[RAW_PROFILES][ctx.config.profile_name].get("outputs", [])],
    key=PROFILE_SELECTOR,
)
# st.sidebar.button("Reload dbt project", key=DBT_DO_RELOAD)
st.sidebar.caption(
    "Refresh the page to reparse dbt. This is useful if any updated models or macros in your physical project \
    on disk have changed and are not yet reflected in the playground as refable or updated."
)

# IDE LAYOUT
notificationContainer = st.empty()
descriptionContainer = st.container()
compileOptionContainer = st.container()
ideContainer = st.container()

descriptionContainer.markdown("")

if not state[PIVOT_LAYOUT]:
    idePart1, idePart2 = ideContainer.columns(2)
else:
    idePart1 = ideContainer.container()
    idePart2 = ideContainer.container()

compileOptionContainer.write("")
auto_update = compileOptionContainer.checkbox("Dynamic Compilation", key=DYNAMIC_COMPILATION, value=True)
if auto_update:
    compileOptionContainer.caption("👉 Compiling SQL on change")
else:
    compileOptionContainer.caption("👉 Compiling SQL with control + enter")

with idePart1:
    state[RAW_SQL] = st_ace(
        value=state[RAW_SQL],
        theme="dracula",  # clouds for white
        language=DBT_ADAPTER_ACE_DIALECT_MAP.get(dbt_profile.Credentials.type, DEFAULT_ACE_DIALECT),
        auto_update=auto_update,
        key="sql",
        max_lines=35,
        min_lines=20,
        height=500,
    )

    st.caption("Request")

    state[FLIGHT_REQUEST] = st_ace(
        value=state[FLIGHT_REQUEST],
        theme="dracula",  # clouds for white
        language="json",
        auto_update=auto_update,
        key="body",
        max_lines=35,
        min_lines=20,
        height=200,
    )

with idePart2:
    with st.expander("📝 Compiled SQL", expanded=True):
        st.code(
            state[COMPILED_SQL]
            if state[COMPILED_SQL]
            else " --> Invalid Jinja, awaiting model to become valid",
            language="sql",
        )


def get_dbt_request_context() -> DbtQueryRequestContext:
    return DbtQueryRequestContext(**json.loads(state[FLIGHT_REQUEST]))


request_context = get_dbt_request_context()
raw_sql_compiled = compile_sql(state[RAW_SQL], request_context)
if raw_sql_compiled != state[COMPILED_SQL]:
    state[COMPILED_SQL, request_context] = raw_sql_compiled
    st.experimental_rerun()  # This eager re-run speeds up the app

if ctx.config.target_name != state[TARGET_PROFILE]:  # or state[DBT_DO_RELOAD]:
    print("Reloading dbt project...")
    with notificationContainer:
        ctx.config.target_name = state[TARGET_PROFILE]
        ctx.config.target_name = state[TARGET_PROFILE]
        with st.spinner("Reloading dbt... ⚙️"):
            inject_dbt(state[TARGET_PROFILE])
            state[COMPILED_SQL] = compile_sql(state[RAW_SQL], get_dbt_request_context())
    st.experimental_rerun()

# TEST LAYOUT
testHeaderContainer = st.container()
test_column_1, _, test_column_2 = st.columns([1, 2, 1])
testContainer = st.container()
testContainerViewer = testContainer.expander("Result Viewer 🔎", expanded=True)
test_view_1, _, test_view_2 = testContainerViewer.columns([1, 2, 1])

downloadBtnContainer, profileBtnContainer = st.columns([1, 1])
profilerContainer = st.container()

with testHeaderContainer:
    st.subheader("🔬 Query Result Inspector")

query_limit = test_column_2.number_input(
    "Limit Results", min_value=1, max_value=50_000, value=2_000, step=1000, key=QUERY_LIMITER,
    help="Limit the number of results returned by the query, the maximum value is 50,000"
)

test_column_1.button(
    "Execute Compiled Query",
    on_click=run_query,
    kwargs={"sql": state[COMPILED_SQL], "limit": query_limit},
)

with testContainerViewer:
    st.write("\n\n\n\n\n")

    if state[SQL_QUERY_STATE] == "success":
        test_view_1.write("#### Compiled SQL query results")
    elif state[SQL_QUERY_STATE] == "error":
        test_view_1.warning(f"SQL query error: {state[SQL_ADAPTER_RESP]}")
    if not state[SQL_RESULT].empty:
        test_view_2.info(f"Adapter Response: {state[SQL_ADAPTER_RESP]}")
        st.dataframe(state[SQL_RESULT])
    else:
        st.write("")
        st.markdown(
            "> The results of your playground query will show up here. Click `Execute Compiled Query` to see the results. "
        )
        st.write("")
    st.write("")

if state[SQL_QUERY_STATE] == "success":
    with downloadBtnContainer:
        st.download_button(
            label="Download data as CSV",
            data=state[SQL_RESULT].to_csv().encode("utf-8"),
            file_name=f"jinjat_playground.csv",
            mime="text/csv",
        )



import json
import typing
from collections import OrderedDict
from typing import (
    Any,
    List,
    Optional,
    Mapping,
)

import agate
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.graph.manifest import ManifestNode
from jsonschema import validate
from openapi_schema_pydantic import OpenAPI
from openapi_schema_pydantic import Operation
from pydantic import BaseModel, validator
from starlette.requests import Request

LIMIT_QUERY_PARAM = '_limit'
JSON_COLUMNS_QUERY_PARAM = '_json_columns'


class CORS(BaseModel):
    allowed_origins: Optional[str]


class JinjatProjectConfig(BaseModel):
    cors: Optional[CORS] = CORS(allowed_origins="*")
    max_limit: Optional[int] = 50000
    default_limit: Optional[int] = 500
    refine: Optional[dict]
    openapi: Optional[dict]

    @validator('openapi')
    def validate_openapi(cls, openapi):
        info = openapi.get("info", {})
        if "title" in info is not None or "version" in info is not None:
            raise ValueError(
                "`openapi.info.title` and `openapi.info.version` must not be set as the values are derived from dbt_project.yml")

        openapi['info']['title'] = openapi['info']['version'] = ""
        OpenAPI.parse_obj(openapi)
        return openapi


class DbtQueryRequestContext(BaseModel):
    method: str
    body: Optional[dict]
    headers: Mapping[str, str] = {}
    params: Mapping[str, Any] = {}
    query: Mapping[str, str] = {}

    def is_debug_enabled(self) -> bool:
        return self.query.get('_debug') is not None


class JinjatAnalysisConfig(BaseModel):
    cors: Optional[bool]
    openapi: Optional[Operation] = Operation()
    method: Optional[str]
    fetch: Optional[bool] = True
    request_model: Optional[str]

    transform_response: Optional[str]
    transform_request: Optional[str]


async def generate_dbt_context_from_request(request: Request, openapi: dict,
                                            transform_request: typing.Callable[[dict], dict]):
    if request.method in ['PATCH', 'PUT', 'POST']:
        body = transform_request(await request.json())
        validate(instance=body, schema=openapi)
    else:
        body = None
    return DbtQueryRequestContext(method=request.method, body=body,
                                  headers=request.headers,
                                  params=request.path_params, query=request.query_params)


class DbtAdapterExecutionResult:
    """Interface for execution results, this keeps us 1 layer removed from dbt interfaces which may change"""

    def __init__(
            self, adapter_response: AdapterResponse, table: agate.Table, raw_sql: str, compiled_sql: str
    ) -> None:
        self.adapter_response = adapter_response
        self.table = table
        self.raw_sql = raw_sql
        self.compiled_sql = compiled_sql


def _convert_table_to_dict(table: agate.Table, json_columns: List[str]):
    output = []
    json_funcs = [(lambda x: json.loads(col.jsonify(x))) if table.column_names[i] in json_columns else col.jsonify
                  for i, col in enumerate(table.column_types)]
    for row in table.rows:
        values = tuple(json_funcs[i](d) for i, d in enumerate(row))
        output.append(OrderedDict(zip(row.keys(), values)))
    return output


class JinjatAdapterResponse(BaseModel):
    message: str
    code: Optional[str] = None
    rows_affected: Optional[int] = None


class JinjatColumn(BaseModel):
    name: str
    type: str


class JinjatExecutionResult(BaseModel):
    """Interface for execution results, this keeps us 1 layer removed from dbt interfaces which may change"""
    request: DbtQueryRequestContext
    adapter_response: JinjatAdapterResponse
    columns: List[JinjatColumn]
    data: Any
    raw_sql: str
    compiled_sql: str

    @staticmethod
    def from_dbt(ctx: DbtQueryRequestContext, result: DbtAdapterExecutionResult,
                 transform_response: typing.Callable[[dict], dict] = None,
                 json_columns: List[str] = []) -> 'JinjatExecutionResult':
        columns = [JinjatColumn(name=column.name, type=column.data_type.__class__.__name__) for column in
                   result.table.columns]
        adapter_response = JinjatAdapterResponse(message=result.adapter_response._message,
                                                 code=result.adapter_response.code,
                                                 rows_affected=result.adapter_response.rows_affected)
        result_dict = _convert_table_to_dict(result.table, json_columns)
        if transform_response is not None:
            result_dict = transform_response(result_dict)
        return JinjatExecutionResult(request=ctx, adapter_response=adapter_response, columns=columns,
                                     data=result_dict, raw_sql=result.raw_sql,
                                     compiled_sql=result.compiled_sql)


class DbtAdapterCompilationResult:
    """Interface for compilation results, this keeps us 1 layer removed from dbt interfaces which may change"""

    def __init__(
            self,
            raw_sql: str,
            compiled_sql: str,
            node: ManifestNode,
            injected_sql: Optional[str] = None,
    ) -> None:
        self.raw_sql = raw_sql
        self.compiled_sql = compiled_sql
        self.node = node
        self.injected_sql = injected_sql



import functools
import multiprocessing
import os
import subprocess
import sys
from pathlib import Path
from typing import Callable, Optional

import click
import uvicorn
from dbt.cli.option_types import YAML
from fal_serverless import sync_dir

from jinjat.core.dbt.config import DEFAULT_PROFILES_DIR
from jinjat.core.generator import compile_macro
from jinjat.core.log_controller import logger
from jinjat.core.server import DbtTarget, SERVER_OPT, app

CONTEXT = {"max_content_width": 800}


@click.group()
@click.version_option()
def cli():
    pass


def shared_server_opts(func: Callable) -> Callable:
    @click.option(
        "--host",
        type=click.STRING,
        help="The host to serve the server on",
        envvar="JINJAT_HOST",
        default="localhost",
    )
    @click.option(
        "--port",
        type=click.INT,
        envvar="JINJAT_PORT",
        help="The port to serve the server on",
        default=8581,
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def serve_project_opts(func: Callable) -> Callable:
    @click.option(
        "--refine",
        is_flag=True,
        help="Enable Refine UI",
        default=False
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def shared_single_project_opts(func: Callable) -> Callable:
    @click.option(
        "--project-dir",
        type=click.Path(exists=True, dir_okay=True, file_okay=False),
        default=str(Path.cwd()),
        help="Which directory to look in for the dbt_project.yml file. Default is the current working directory and its parents.",
    )
    @click.option(
        "--profiles-dir",
        type=click.Path(exists=True, dir_okay=True, file_okay=False),
        default=DEFAULT_PROFILES_DIR,
        help="Which directory to look in for the profiles.yml file. Defaults to ~/.dbt",
    )
    @click.option(
        "-t",
        "--target",
        type=click.STRING,
        help="Which profile to load. Overrides setting in dbt_project.yml.",
    )
    @click.option(
        "--vars",
        envvar=None,
        default='{}',
        help="Supply variables to the project. This argument overrides variables defined in your dbt_project.yml file. This argument should be a YAML string, eg. '{my_variable: my_value}'",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@cli.command(context_settings=CONTEXT)
@shared_single_project_opts
@click.argument(
    "macro",
    type=click.STRING,
)
@click.option(
    "--args",
    type=YAML(),
    default="{}",
    help="""Supply arguments to the macro. This dictionary will be mapped to the
            keyword arguments defined in the selected macro. This argument should
            be a YAML string, eg. '{my_variable: my_value}'
            """,
)
@click.option(
    "--dry-run",
    is_flag=True
)
def generate(
        macro: str,
        args: dict,
        dry_run: bool,
        project_dir: str,
        profiles_dir: str,
        target: Optional[str],
        vars: str
):
    dbt_target = DbtTarget(project_dir=project_dir, profiles_dir=profiles_dir, target=target, vars=vars)
    compile_macro(dbt_target, macro, args, dry_run)

@serve_project_opts
@cli.command(context_settings=CONTEXT)
@shared_single_project_opts
@shared_server_opts
def serve(
        project_dir: str,
        profiles_dir: str,
        target: Optional[str],
        host: str,
        port: int,
        vars: str,
        refine: bool,
) -> object:
    logger().info(":water_wave: Executing jinjat in single-tenant mode")

    dbt_target = DbtTarget(project_dir=project_dir, profiles_dir=profiles_dir, target=target, vars=vars, refine=refine)
    os.environ[SERVER_OPT] = dbt_target.json()

    server = multiprocessing.Process(target=run_server, args=(host, port))
    server.start()

    import atexit

    atexit.register(lambda: server.terminate())

    server.join()
    sys.exit(server.exitcode)


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@shared_single_project_opts
@shared_server_opts
@click.pass_context
def streamlit(
        ctx,
        project_dir: str,
        profiles_dir: str,
        target: str,
        host: str,
        port: int,
):
    """Start the jinjat playground
    \f
    Pass the --options command to see streamlit specific options that can be passed to the app,
    pass --config to see the output of streamlit config show
    """

    logger().info(":water_wave: Executing Jinjat Playground\n")

    if "--options" in ctx.args:
        subprocess.run(["streamlit", "run", "--help"])
        ctx.exit()

    import os

    if "--config" in ctx.args:
        subprocess.run(
            ["streamlit", "config", "show"],
            env=os.environ,
            cwd=Path.cwd(),
        )
        ctx.exit()

    script_args = ["--"]
    if project_dir:
        script_args.append("--project-dir")
        script_args.append(project_dir)
    if profiles_dir:
        script_args.append("--profiles-dir")
        script_args.append(profiles_dir)
    if target:
        script_args.append("--target")
        script_args.append(target)

    streamlit_command = ["streamlit", "run", "--runner.magicEnabled=false",
                         Path(__file__).parent / "playground.py", ] + ctx.args + script_args
    print(streamlit_command)
    subprocess.run(
        streamlit_command,
        env=os.environ,
        cwd=Path.cwd(),
    )


def run_server(host="localhost", port=8581):
    app.state.test = 1
    uvicorn.run(
        "jinjat.core.server:app",
        host=host,
        port=port,
        log_level="info",
        reload=False,
        workers=1,
    )


if __name__ == "__main__":
    cli()



from typing import (
    Optional, TypeVar,
)
from dbt.flags import DEFAULT_PROFILES_DIR
from dbt.tracking import disable_tracking
from dbt.version import __version__ as dbt_version
from ruamel.yaml import YAML

CACHE = {}
CACHE_VERSION = 1
JINJAT_REQUEST_VAR_NAME = "jinjat_request"

DBT_MAJOR_VER, DBT_MINOR_VER, DBT_PATCH_VER = (int(v) for v in dbt_version.split("."))
RAW_CODE = "raw_code" if DBT_MAJOR_VER >= 1 and DBT_MINOR_VER >= 3 else "raw_sql"
COMPILED_CODE = "compiled_code" if DBT_MAJOR_VER >= 1 and DBT_MINOR_VER >= 3 else "compiled_sql"

JINJA_CONTROL_SEQS = ["{{", "}}", "{%", "%}", "{#", "#}"]
T = TypeVar("T")


def has_jinja(query: str) -> bool:
    """Utility to check for jinja prior to certain compilation procedures"""
    return any(seq in query for seq in JINJA_CONTROL_SEQS)


disable_tracking()
fire_event = lambda e: None


class ConfigInterface:
    """This mimic dbt-core args based interface for dbt-core
    class instantiation"""

    def __init__(
            self,
            threads: Optional[int] = 1,
            target: Optional[str] = None,
            profiles_dir: Optional[str] = None,
            project_dir: Optional[str] = None,
            vars: Optional[str] = "{}",
            profile: Optional[str] = None,
    ):
        self.threads = threads
        self.target = target
        self.profiles_dir = profiles_dir or DEFAULT_PROFILES_DIR
        self.project_dir = project_dir
        self.vars = vars  # json.dumps str
        self.dependencies = []
        self.single_threaded = threads == 1
        self.quiet = True
        self.profile = profile

    @classmethod
    def from_str(cls, arguments: str) -> "ConfigInterface":
        import argparse
        import shlex

        parser = argparse.ArgumentParser()
        args = parser.parse_args(shlex.split(arguments))
        return cls(
            threads=args.threads,
            target=args.target,
            profiles_dir=args.profiles_dir,
            project_dir=args.project_dir,
        )


class YamlHandler(YAML):
    """A `ruamel.yaml` wrapper to handle dbt YAML files with sane defaults"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.indent(mapping=2, sequence=4, offset=2)
        self.width = 800
        self.preserve_quotes = True
        self.default_flow_style = False



from dataclasses import dataclass
from enum import Enum
from typing import Any, List

from fastapi.exceptions import HTTPException
from pydantic import BaseModel, ValidationError
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import HTMLResponse, Response, JSONResponse


DBT_PROJECT_HEADER = 'x-dbt-project-version'
DBT_PROJECT_NAME = 'x-dbt-project-name'

def get_human_readable_error(validation_error: ValidationError) -> str:
    error_str = "Validation errors:\n"
    for error in validation_error.errors():
        error_str += f'-> {error.get("loc")}: {error.get("msg")}\n'
    return error_str


class JinjatErrorCode(int, Enum):
    FailedToReachServer = -1
    CompileSqlFailure = 1
    ExecuteSqlFailure = 2
    ProjectParseFailure = 3
    ProjectNotFound = 4
    ProjectHeaderNotSupplied = 5
    SqlNotSupplied = 6
    JmesPathParseError = 7


class JinjatError(BaseModel):
    code: JinjatErrorCode
    error: Any


class JinjatErrorContainer(HTTPException):
    def __init__(self, status_code: int, error: JinjatError) -> None:
        super().__init__(status_code, error.dict())


@dataclass
class CustomButton:
    name: str
    url: str

async def rapidoc_html(custom_button: CustomButton, request: Request) -> HTMLResponse:
    html = f"""
        <!doctype html>
        <html>
            <head>
                <meta charset="utf-8">
                <script 
                    type="module" 
                    src="https://unpkg.com/rapidoc/dist/rapidoc-min.js"
                ></script>
            </head>
            <body>
                <rapi-doc spec-url="openapi.json" 
                          primary-color="#DCBB0E" render-style="focused" 
                          show-method-in-nav-bar="as-colored-block" 
                          show-header="false" 
                          show-components="true"
                          show-curl-before-try="true">
                    <div slot="overview">
                        <a href="openapi.json" style="color:#DCBB0E">openapi.json</a><br>
                    </div>
                    <div slot="nav-logo">
                        <img src="https://jinjat.com/img/jinjat-logo.svg" style="height:35px;margin-left: 7px;">
                        {f'<a href="{custom_button.url}" style="background: #DCBB0E;padding: 0 8px;border-radius: 3px;font-weight: bold;color: #000;text-decoration: none;line-height: 35px;height: 35px;float: right;margin-right: 7px;">{custom_button.name}</a>' if custom_button is not None else ""}
                    </div>
                </rapi-doc>
            </body> 
        </html>
    """
    return HTMLResponse(html)

async def elements_html(custom_button: CustomButton, request: Request) -> HTMLResponse:
    html = f"""
        <!doctype html>
        <html lang="en">
          <head>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <title>Elements in HTML</title>
          
            <script src="https://unpkg.com/@stoplight/elements/web-components.min.js"></script>
            <link rel="stylesheet" href="https://unpkg.com/@stoplight/elements/styles.min.css">
          </head>
          <body>
        
            <elements-api
              apiDescriptionUrl="openapi.json"
              layout="sidebar"
              logo="https://jinjat.com/img/jinjat-logo.svg"
              router="history"
            />
        
          </body>
        </html>

    """
    return HTMLResponse(html)


class JSONAPIResponse(JSONResponse):
    """
    Base response class for json:api requests, sets `Content-Type: application/vnd.api+json`.
    For detailed information, see `Starlette responses <https://www.starlette.io/responses/>`_.
    """
    media_type = 'application/vnd.api+json'

    def render(self, content: Any) -> bytes:
        if content is None:
            return b''
        return super().render(content)


class JSONAPIException(HTTPException):
    """ HTTP exception with json:api representation. """

    def __init__(self, status_code: int, detail: str = None, errors: List[dict] = None) -> None:
        """
        Base json:api exception class.
        :param status_code: HTTP status code
        :param detail: Optional, error detail, will be serialized in the final HTTP response.
                       **DO NOT** include sensitive information here.
                       If not specified, the HTTP message associated to ``status_code``
                       will be used.
        :param errors: Optional, list of json:api error representations.
                       Used if multiple errors are returned.
                       .. code-block:: python
                           import json
                           from starlette_jsonapi.utils import serialize_error
                           error1 = JSONAPIException(400, 'foo')
                           error2 = JSONAPIException(400, 'bar')
                           final_error = JSONAPIException(
                               400, 'final', errors=error1.errors + error2.errors
                           )
                           response = serialize_error(final_error)
                           assert json.loads(response.body)['errors'] == [
                               {'detail': 'foo'},
                               {'detail': 'bar'},
                               {'detail': 'final'},
                           ]
        """
        super().__init__(status_code, detail=detail)
        self.errors = errors or []
        self.errors.append({'detail': self.detail})


def register_jsonapi_exception_handlers(app: Starlette):
    """
    Registers exception handlers on a Starlette app,
    serializing uncaught Exception and HTTPException to a json:api compliant body.
    """

    def serialize_error(exc: Exception) -> JSONAPIResponse:
        """
        Serializes exception according to the json:api spec
        and returns the equivalent :class:`JSONAPIResponse`.
        """
        if isinstance(exc, JinjatErrorContainer):
            status_code = exc.status_code
            errors = [exc.detail]
        if isinstance(exc, JSONAPIException):
            status_code = exc.status_code
            errors = exc.errors
        elif isinstance(exc, HTTPException):
            status_code = exc.status_code
            errors = [{'detail': exc.detail}]
        else:
            status_code = 500
            errors = [{'detail': 'Internal server error'}]

        error_body = {
            'errors': errors
        }
        return JSONAPIResponse(status_code=status_code, content=error_body)

    async def _serialize_error(request: Request, exc: Exception) -> Response:
        return serialize_error(exc)

    app.add_exception_handler(Exception, _serialize_error)
    app.add_exception_handler(HTTPException, _serialize_error)





# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jinjat',
 'jinjat.core',
 'jinjat.core.dbt',
 'jinjat.core.routes',
 'jinjat.core.schema',
 'jinjat.core.util']

package_data = \
{'': ['*']}

install_requires = \
['bottle>=0.12.23,<0.13.0',
 'click>7',
 'dbt-core==1.4.5',
 'deepmerge>=1.1.0,<2.0.0',
 'fastapi>=0.85.0,<0.86.0',
 'jinja2-simple-tags>=0.4.0,<0.5.0',
 'jmespath>=1.0.1,<2.0.0',
 'jsonref>=1.1.0,<2.0.0',
 'jsonschema>=3.0,<4.0',
 'openapi-schema-pydantic>=1.2.4,<2.0.0',
 'orjson>=3.8.0,<4.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'rich>=10',
 'ruamel.yaml>=0.17',
 'uvicorn[standard]>=0.18.3,<0.19.0',
 'watchdog>=2.2.1']

extras_require = \
{'duckdb': ['duckcli>=0.2.1,<0.3.0', 'dbt-duckdb>=1.4.1,<2.0.0'],
 'playground': ['streamlit>=1.0.0',
                'streamlit-ace>=0.1.0',
                'graphviz>=0.17',
                'pydot>=1.4.2',
                'streamlit-agraph>=0.0.35',
                'streamlit-pandas-profiling>=0.1.3',
                'streamlit-aggrid>=0.2.2',
                'scipy>=1.3.1,<2.0.0',
                'feedparser>=6.0.10,<7.0.0']}

entry_points = \
{'console_scripts': ['jinjat = jinjat.main:cli']}

setup_kwargs = {
    'name': 'jinjat',
    'version': '0.3',
    'description': 'A low-code data application framework that uses dbt Core and OpenAPI',
    'long_description': '# Jinjat\n\n## Develop data applications with dbt, SQL, and OpenAPI\n\n### Installation\n\n```commandline\npip install jinjat\n```\n\n### Create your first API\n\nCreate an [analysis]() in `analysis/my_first_api.sql`:\n```sql\n{%- set query = request().query %}\n\nselect \'{{query.example}}\' as col1\n```\n\nAnd create a YML file in `analysis/schema.yml`:\n\n```yml\nversion: 2\n\nanalyses:\n  - name: my_first_api\n    config:\n      jinjat:\n        method: get\n        openapi:\n          parameters:\n            - in: query\n              name: example\n              schema:\n                type: number\n```\n\nStart Jinjat as follows:\n\n```commandline\njinjat serve --project-dir [YOUR_DBT_PROJECT_DIRECTORY]\n```\n\nAnd then run the following CURL command to test the API:\n\n```commandline\ncurl -XGET \'http://127.0.0.1:8581?example=value\'\n```\n\nIt should return the following response:\n\n```json\n[\n  "col1": "3"\n]\n```\n\nJinjat uses OpenAPI to validate the requests and create an API documentation automatically for your API.\n\n## Integrations\n\npoetry install --extras "duckdb"\n\n### Playground\n\npoetry install --extras "playground"\n\n\n#### Installation\n\n```commandline\npip install jinjat[playground]\n```\n\nJinjat Playground is a Streamlit app that lets you develop APIs in your browser.\nOnce you write the template, you can save it to your dbt project as an analysis and expose the API.\n\n### [Refine.dev](https://refine.dev) Integration\n\n#### Installation\n\n```commandline\npip install jinjat[refine]\n```\n\nJinjat Refine integration creates a Refine app from your OpenAPI spec \n\n\n> Jinjat is a fork of [dbt-osmosis](https://github.com/z3z1ma/dbt-osmosis)',
    'author': 'buremba',
    'author_email': 'emrekabakci@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jinjat-data/jinjat',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



import asyncio
import functools
import json
from datetime import datetime
from typing import Optional, Union

from fastapi import FastAPI
from pydantic import BaseModel, Field
from starlette import status
from starlette.background import BackgroundTasks
from starlette.requests import Request
from starlette.responses import Response

from jinjat.core.dbt.dbt_project import DbtProjectContainer, DbtProject
from jinjat.core.exceptions import ExecuteSqlFailure
from jinjat.core.models import JinjatExecutionResult, DbtAdapterExecutionResult, generate_dbt_context_from_request, \
    DbtQueryRequestContext, JSON_COLUMNS_QUERY_PARAM
from jinjat.core.util.api import JinjatErrorContainer, JinjatError, JinjatErrorCode, DBT_PROJECT_HEADER, \
    DBT_PROJECT_NAME
from jinjat.core.util.jmespath import extract_jmespath

app = FastAPI(redoc_url=None, docs_url=None, title="Admin API", version="0.1")

class JinjatCompileResult(BaseModel):
    result: str


class JinjatRefreshProjectResult(BaseModel):
    result: str


def jinjat_project_not_found_error():
    return JinjatErrorContainer(
        status_code=status.HTTP_400_BAD_REQUEST,
        error=JinjatError(code=JinjatErrorCode.ProjectNotFound, error="Project could not be found"))


class DbtAdhocQueryRequest(BaseModel):
    sql: str = Field(examples=["select 1"])
    request: DbtQueryRequestContext
    limit: Optional[int]


@app.get(
    "/manifest.json"
)
async def execute_manifest_query(
        request: Request,
        response: Response,
        jmespath: Optional[str] = None,
        # x_dbt_project: Optional[str] = Header(default=None),
) -> any:
    """Execute dbt SQL against a registered project as determined by X-dbt-Project header"""
    dbt_container: DbtProjectContainer = request.app.state.dbt_project_container
    project = dbt_container.get_project(None)

    if project is None:
        raise jinjat_project_not_found_error()


    manifest_file = project.dbt.writable_manifest().to_dict()
    result = extract_jmespath(jmespath, manifest_file, project)
    response.headers[DBT_PROJECT_HEADER] = project.config.version
    response.headers[DBT_PROJECT_NAME] = project.config.project_name
    return result


@app.post(
    "/execute",
    response_model=JinjatExecutionResult
)
async def execute_sql(
        request: Request,
        body: DbtAdhocQueryRequest,
        # x_dbt_project: Optional[str] = Header(default=None),
) -> JinjatExecutionResult:
    """Execute dbt SQL against a registered project as determined by X-dbt-Project header"""
    dbt: DbtProjectContainer = request.app.state.dbt_project_container
    project = dbt.get_project(None)

    if project is None:
        raise jinjat_project_not_found_error()

    dbt_result = await _execute_jinjat_query(project, project.execute_sql, body.sql, body.request, body.limit)
    json_cols = json.loads(request.query_params.get(JSON_COLUMNS_QUERY_PARAM) or '[]')

    return JinjatExecutionResult.from_dbt(body.request, dbt_result, json_columns=json_cols)


async def _execute_jinjat_query(project: DbtProject, execute_function, query: str, ctx: DbtQueryRequestContext,
                                limit: Optional[int], fetch: bool = True) -> DbtAdapterExecutionResult:
    if limit is not None:
        query = project.execute_macro('limit_query', {"sql": query, "limit": limit})

    try:
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(
            None, project.fn_threaded_conn(execute_function, query, ctx, fetch)
        )
    except ExecuteSqlFailure as execution_err:
        raise JinjatErrorContainer(
            status_code=status.HTTP_400_BAD_REQUEST,
            error=JinjatError(
                code=JinjatErrorCode.ExecuteSqlFailure,
                error=execution_err.to_model()
            )
        )

    return result


@app.post(
    "/compile",
    response_model=JinjatCompileResult
)
async def compile_sql(
        request: Request,
        body: DbtAdhocQueryRequest,
        # x_dbt_project: str = Header(),
) -> JinjatCompileResult:
    """Compile dbt SQL against a registered project as determined by X-dbt-Project header"""
    dbt: DbtProjectContainer = request.app.state.dbt_project_container

    project = dbt.get_project(None)
    if project is None:
        raise jinjat_project_not_found_error()

    # Query Compilation

    try:
        loop = asyncio.get_running_loop()

        if body.limit is not None:
            query = project.execute_macro('limit_query', {"sql": body.sql, "limit": body.limit})
        else:
            query = body.sql

        context = generate_dbt_context_from_request(request) if body.request is None else body.request
        compiled_query = (
            await loop.run_in_executor(
                None,
                project.fn_threaded_conn(project.compile_sql, query, context)
            )
        ).compiled_sql
    except Exception as compile_err:
        raise JinjatErrorContainer(
            status_code=status.HTTP_400_BAD_REQUEST,
            error=JinjatError(
                code=JinjatErrorCode.CompileSqlFailure,
                error=str(compile_err),
            )
        )

    return JinjatCompileResult(result=compiled_query)


@app.get(
    "/refresh",
)
async def refresh(
        background_tasks: BackgroundTasks,
        request: Request,
        response: Response,
        reset: bool = False,
        target: Optional[str] = None,
        # x_dbt_project: str = Header(),
) -> JinjatRefreshProjectResult:
    """Reparse a registered project on disk as determined by X-dbt-Project header writing
    manifest.json to target directory"""
    dbt: DbtProjectContainer = request.app.state.dbt_project_container
    project = dbt.get_project(None)
    if project is None:
        raise jinjat_project_not_found_error()

    # Get targets
    old_target = getattr(project.args, "target", project.config.target_name)
    new_target = target or old_target

    if not reset and old_target == new_target:
        # Async (target same)
        if project.mutex.acquire(blocking=False):
            background_tasks.add_task(_reset, project, reset, old_target, new_target)
            return JinjatRefreshProjectResult(result="Initializing project parsing")
        else:
            return JinjatRefreshProjectResult(result="Currently re-parsing project")
    else:
        # Sync (target changed or reset is true)
        if project.mutex.acquire(blocking=old_target != new_target):
            rv = _reset(project, reset, old_target, new_target)
            if isinstance(rv, JinjatErrorContainer):
                response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
            return rv
        else:
            return JinjatRefreshProjectResult(result="Currently re-parsing project")


def _reset(
        project: DbtProject, reset: bool, old_target: str, new_target: str
) -> Union[JinjatRefreshProjectResult, JinjatErrorContainer]:
    """Use a mutex to ensure only a single reset can be running for any
    given project at any given time synchronously or asynchronously"""
    target_did_change = old_target != new_target
    try:
        project.args.target = new_target
        project.safe_parse_project(reinit=reset or target_did_change)
    except Exception as reparse_err:
        project.args.target = old_target
        rv = JinjatErrorContainer(
            status_code=status.HTTP_400_BAD_REQUEST,
            error=JinjatError(
                code=JinjatErrorCode.ProjectParseFailure,
                error=str(reparse_err),
            )
        )
    else:
        project._version += 1
        rv = JinjatRefreshProjectResult(
            result=(
                f"Profile target changed from {old_target} to {new_target}!"
                if target_did_change
                else f"Reparsed project with profile {old_target}!"
            )
        )
    finally:
        project.mutex.release()
    return rv


@app.get("/health")
async def health_check(
        request: Request,
        # x_dbt_project: str = Header(),
) -> dict:
    """Checks if the server is running and accepting requests"""
    dbt: DbtProjectContainer = request.app.state.dbt_project_container
    project = dbt.get_project(None)

    return {
        "result": {
            "status": "ready",
            **(
                {
                    "project_name": project.config.project_name,
                    "target_name": project.config.target_name,
                    "profile_name": project.config.profile_name,
                    "logs": project.config.log_path,
                    "runner_parse_iteration": project._version,
                    "adapter_ready": project.adapter_probe(),
                }
                if project is not None
                else {}
            ),
            "timestamp": datetime.utcnow(),
            "error": None,
        },
        "jinjat": __name__,
    }



from typing import Optional

from pydantic import BaseModel

class DbtExecutionError(BaseModel):
    raw_sql: str
    compiled_sql: Optional[str]
    error: str

class InvalidJinjaConfig(SystemExit):

    def __init__(self, schema_file_path: str, sql_file_path: str, *args: object) -> None:
        super().__init__(*args)
        self.schema_file_path = schema_file_path
        self.sql_file_path = sql_file_path


class ExecuteSqlFailure(RuntimeError):

    def __init__(self, raw_sql: str, compiled_sql: Optional[str], dbt_exception: Exception):
        self.raw_sql = raw_sql
        self.compiled_sql = compiled_sql
        self.dbt_exception = dbt_exception

    def to_model(self) -> DbtExecutionError:
        return DbtExecutionError(raw_sql=self.raw_sql, compiled_sql=self.compiled_sql, error=str(self.dbt_exception))

class SanitizationRequired(Exception):
    pass



from typing import Optional, List

import jmespath
from dbt.exceptions import CompilationError
from jmespath.exceptions import ParseError, JMESPathError
from starlette import status

from jinjat.core.dbt.dbt_project import DbtProject
from jinjat.core.util.api import JinjatErrorContainer, JinjatError, JinjatErrorCode


class CustomFunctions(jmespath.functions.Functions):

    @jmespath.functions.signature(
        {'types': ['string'], "variadic": True})
    def _func_projection(self, picked_keys: str, *values):
        picked = [key.strip() for key in picked_keys.split(',')]
        ret = {}
        for i, key in enumerate(picked):
            ret[key] = values[i]
        return ret


def json_query(data, expr):
    '''Query data using jmespath query language ( http://jmespath.org ). Example:
    - debug: msg="{{ instance | json_query(tagged_instances[*].block_device_mapping.*.volume_id') }}"
    '''

    options = jmespath.Options(custom_functions=CustomFunctions())

    return jmespath.search(expr, data, options=options)


class FilterModule(object):
    ''' Query filter '''

    def filters(self):
        return {
            'json_query': json_query
        }


def extract_jmespath(query: Optional[str], data: dict, dbt : DbtProject):
    current_data = data
    if query is None or query == '' or query == '*':
        return current_data

    try:
        compiled_expression = dbt.compile_sql(query)
    except CompilationError as e:
        raise JinjatErrorContainer(
            status_code=status.HTTP_400_BAD_REQUEST,
            error=JinjatError(code=JinjatErrorCode.JmesPathParseError, error=f"{e}"))

    try:
        jmespath_compile = jmespath.compile(compiled_expression.compiled_sql)
        current_data = jmespath_compile.search(current_data, jmespath.Options(custom_functions=CustomFunctions()))
    except JMESPathError as e:
        raise JinjatErrorContainer(
            status_code=status.HTTP_400_BAD_REQUEST,
            error=JinjatError(code=JinjatErrorCode.JmesPathParseError, error=f"{e}"))
    return current_data
